import textwrap

import pytest

from .parser import Issue, parse


@pytest.mark.parametrize(
    "fragment, expected, description_expected",
    [
        ("", Issue(), False),
        ("a simple test right there", Issue(title="a simple test right there"), False),
        (
            "@me some ~labels to ~organize issues ~bug",
            Issue(
                title="some labels to organize issues",
                labels={"labels", "organize", "bug"},
                assignees={"me"},
            ),
            False,
        ),
        (
            "a %milestone to keep ~track of stuff",
            Issue(
                title="a milestone to keep track of stuff",
                labels={"track"},
                milestone="milestone",
            ),
            False,
        ),
        (
            "A label with a description following it ~now:",
            Issue(title="A label with a description following it", labels={"now"}),
            True,
        ),
        (
            "#.1 An issue with a reference definition",
            Issue(title="An issue with a reference definition", reference=1),
            False,
        ),
    ],
)
def test_parse_fragment(fragment, expected, description_expected):
    actual, expecting_description = Issue.parse(fragment)
    assert expecting_description == description_expected
    assert actual == expected


@pytest.mark.parametrize(
    "lines, expected",
    [
        ("", []),
        ("A simple issue", [Issue(title="A simple issue")]),
        ("~label @me", []),
        (
            """
            @me some ~labels to ~organize issues ~bug
            a %milestone to keep ~track of stuff
            """,
            [
                Issue(
                    title="some labels to organize issues",
                    labels={"labels", "organize", "bug"},
                    assignees={"me"},
                ),
                Issue(
                    title="a milestone to keep track of stuff",
                    labels={"track"},
                    milestone="milestone",
                ),
            ],
        ),
        (
            """
            some stuff
            \tinside: not processed
            """,
            [
                Issue(title="some stuff"),
            ],
        ),
        (
            """
            ~common-tag @someone
            \tright there ~other-tag
            \t//A comment

            \t@someone-else right %here
            """,
            [
                Issue(
                    title="right there",
                    labels={"common-tag", "other-tag"},
                    assignees={"someone"},
                ),
                Issue(
                    title="right",
                    labels={"common-tag"},
                    assignees={"someone-else", "someone"},
                    milestone="here",
                ),
            ],
        ),
        (
            """An ~issue with a description:
\tThis is the %description of the issue:
\t// This is *not* a comment
\tIt has a 
\t- bullet list

\tAnd
\t\tIndentation
            """,
            [
                Issue(
                    title="An issue with a description",
                    labels={"issue"},
                    description="""This is the %description of the issue:
// This is *not* a comment
It has a
- bullet list
And
\tIndentation
""",
                )
            ],
        ),
        (
            """%milestone_test @me
\t~notsure do this
\t~important do that
""",
            [
                Issue(
                    title="do this",
                    labels={"notsure"},
                    assignees={"me"},
                    milestone="milestone_test",
                ),
                Issue(
                    title="do that",
                    labels={"important"},
                    assignees={"me"},
                    milestone="milestone_test",
                ),
            ],
        ),
        (
            """
An issue that references another ~blocked:
\tSee #.1

#.1 The other one ^w^
""",
            [
                Issue(
                    title="An issue that references another",
                    labels={"blocked"},
                    description="See #.1\n",
                ),
                Issue(
                    title="The other one ^w^",
                    reference=1,
                ),
            ],
        ),
    ],
)
def test_parse_issues(lines, expected):
    assert list(parse(textwrap.dedent(lines))) == expected


def test_resolves_references():
    [issue, *_] = list(parse("An issue that references another ~blocked:\n\tSee #.2"))
    assert issue.references == {2}
    issue = issue.resolve_references({2: 1})
    assert issue.description == "See #1\n"


def test_splits_sigil():
    assert Issue._word_and_sigil("#.1") == ("#.", "1")


def test_parse_issue_with_missing_description_fails():
    with pytest.raises(
        ValueError, match="Expected a description after 'An ~issue with a description:'"
    ):
        list(parse("An ~issue with a description:\nNo description here"))



import os
from pathlib import Path
from unittest.mock import Mock
from urllib.parse import urlparse

import pytest

from issurge.main import run
from issurge.parser import Issue, subprocess
from issurge.utils import debugging, dry_running


class MockedSubprocessOutput:
    def __init__(self, stdout: str, stderr: str):
        self.stdout = stdout.encode("utf-8")
        self.stderr = stderr.encode("utf-8")


@pytest.fixture
def setup():
    Path("test_empty_issues").write_text("")
    Path("test_some_issues").write_text(
        """~common @common %common
\tAn issue to submit
Another ~issue to submit @me"""
    )
    subprocess.run = Mock(
        return_value=MockedSubprocessOutput(
            "https://github.com/gwennlbh/gh-api-playground/issues/5\n",
            "Some unrelated stuff haha",
        )
    )
    Issue._get_remote_url = Mock(
        return_value=urlparse("https://github.com/ewen-lbh/gh-api-playground")
    )
    yield
    Path("test_empty_issues").unlink()
    Path("test_some_issues").unlink()
    del os.environ["ISSURGE_DEBUG"]
    del os.environ["ISSURGE_DRY_RUN"]


@pytest.fixture
def default_opts():
    return {
        "<submitter-args>": [],
        "<file>": "test_empty_issues",
        "<words>": [],
        "new": False,
        "--dry-run": False,
        "--debug": False,
    }


def test_dry_run_is_set_when_dry_run_is_passed(setup, default_opts):
    run(opts={**default_opts, "--dry-run": True})
    assert dry_running()
    assert not debugging()


def test_debug_is_set_when_debug_is_passed(setup, default_opts):
    run(opts={**default_opts, "--debug": True})
    assert debugging()
    assert not dry_running()


def test_dry_run_and_debug_are_not_set_by_default(setup, default_opts):
    run(opts=default_opts)
    assert not dry_running()
    assert not debugging()


def test_both_dry_run_and_debug_are_set_when_both_are_passed(setup, default_opts):
    run(opts={**default_opts, "--dry-run": True, "--debug": True})
    assert dry_running()
    assert debugging()


def test_issues_are_submitted_when_dry_run_is_not_passed_with_github_provider(
    setup, default_opts
):
    run(opts={**default_opts, "<file>": "test_some_issues"})
    assert [call.args[0] for call in subprocess.run.mock_calls] == [
        [
            "gh",
            "issue",
            "new",
            "-t",
            "An issue to submit",
            "-b",
            "",
            "-a",
            "common",
            "-l",
            "common",
            "-m",
            "common",
        ],
        [
            "gh",
            "issue",
            "new",
            "-t",
            "Another issue to submit",
            "-b",
            "",
            "-a",
            "@me",
            "-l",
            "issue",
        ],
    ]


def test_issues_are_submitted_when_dry_run_is_not_passed_with_gitlab_provider(
    setup, default_opts
):
    Issue._get_remote_url = Mock(
        return_value=urlparse("https://gitlab.com/ewen-lbh/gh-api-playground")
    )
    run(opts={**default_opts, "<file>": "test_some_issues"})
    assert [call.args[0] for call in subprocess.run.mock_calls] == [
        [
            "glab",
            "issue",
            "new",
            "-t",
            "An issue to submit",
            "-d",
            "",
            "-a",
            "common",
            "-l",
            "common",
            "-m",
            "common",
        ],
        [
            "glab",
            "issue",
            "new",
            "-t",
            "Another issue to submit",
            "-d",
            "",
            "-a",
            "@me",
            "-l",
            "issue",
        ],
    ]


def test_issues_are_not_submitted_when_dry_run_is_passed(setup, default_opts):
    run(opts={**default_opts, "<file>": "test_some_issues", "--dry-run": True})
    assert len(subprocess.run.mock_calls) == 0


def test_issues_are_not_submitted_when_dry_run_is_passed_in_interactive_mode(
    setup, default_opts
):
    run(
        opts={
            **default_opts,
            "new": True,
            "--dry-run": True,
            "<words>": ["testing", "~this", "issue", "@me"],
        }
    )
    assert len(subprocess.run.mock_calls) == 0


def test_issues_are_submitted_when_dry_run_is_not_passed_in_interactive_mode_github_provider(
    setup, default_opts
):
    run(
        opts={
            **default_opts,
            "new": True,
            "<words>": ["testing", "~this", "issue", "@me"],
        }
    )
    assert [call.args[0] for call in subprocess.run.mock_calls] == [
        [
            "gh",
            "issue",
            "new",
            "-t",
            "testing this issue",
            "-b",
            "",
            "-a",
            "@me",
            "-l",
            "this",
        ],
    ]


def test_issues_are_submitted_when_dry_run_is_not_passed_in_interactive_mode_gitlab_provider(
    setup, default_opts
):
    Issue._get_remote_url = Mock(
        return_value=urlparse("https://gitlab.com/ewen-lbh/gh-api-playground")
    )
    run(
        opts={
            **default_opts,
            "new": True,
            "<words>": ["testing", "~this", "issue", "@me"],
        }
    )
    assert [call.args[0] for call in subprocess.run.mock_calls] == [
        [
            "glab",
            "issue",
            "new",
            "-t",
            "testing this issue",
            "-d",
            "",
            "-a",
            "@me",
            "-l",
            "this",
        ],
    ]



import os
import re
import subprocess
from typing import Any, Iterable, NamedTuple
from urllib.parse import urlparse

from rich import print

from issurge.utils import NEWLINE, TAB, debug, debugging, dry_running


class Node:
    def __init__(self, indented_line):
        self.children = []
        self.level = len(indented_line) - len(indented_line.lstrip())
        self.text = indented_line.strip()

    def add_children(self, nodes):
        childlevel = nodes[0].level
        while nodes:
            node = nodes.pop(0)
            if node.level == childlevel:  # add node as a child
                self.children.append(node)
            elif (
                node.level > childlevel
            ):  # add nodes as grandchildren of the last child
                nodes.insert(0, node)
                self.children[-1].add_children(nodes)
            elif node.level <= self.level:  # this node is a sibling, no more children
                nodes.insert(0, node)
                return

    def as_dict(self) -> dict[str, Any]:
        if len(self.children) > 1:
            child_dicts = {}
            for node in self.children:
                child_dicts |= node.as_dict()
            return {self.text: child_dicts}
        elif len(self.children) == 1:
            return {self.text: self.children[0].as_dict()}
        else:
            return {self.text: None}

    @staticmethod
    def to_dict(to_parse: str) -> dict[str, Any]:
        if not to_parse.strip():
            return {}
        root = Node("root")
        root.add_children(
            [Node(line) for line in to_parse.splitlines() if line.strip()]
        )
        return root.as_dict()["root"]


class Issue(NamedTuple):
    title: str = ""
    description: str = ""
    labels: set[str] = set()
    assignees: set[str] = set()
    milestone: str = ""
    reference: int|None = None

    def __rich_repr__(self):
        yield self.title
        yield "description", self.description, ""
        yield "labels", self.labels, set()
        yield "assignees", self.assignees, set()
        yield "milestone", self.milestone, ""
        yield "ref", self.reference, None
        yield "references", self.references, set()

    def __str__(self) -> str:
        result = ""
        if self.reference:
            result += f"<#{self.reference}> "
        result += f"{self.title}" or "<No title>"
        if self.labels:
            result += f" {' '.join(['~' + l for l in self.labels])}"
        if self.milestone:
            result += f" %{self.milestone}"
        if self.assignees:
            result += f" {' '.join(['@' + a for a in self.assignees])}"
        if self.description:
            result += f": {self.description}"
        return result

    def display(self) -> str:
        result = ""
        if self.reference:
            result += f"[bold blue]<#{self.reference}>[/bold blue] "
        result += f"[white]{self.title[:30]}[/white]" or "[red]<No title>[/red]"
        if len(self.title) > 30:
            result += " [white dim](...)[/white dim]"
        if self.labels:
            result += (
                f" [yellow]{' '.join(['~' + l for l in self.labels][:4])}[/yellow]"
            )
            if len(self.labels) > 4:
                result += " [yellow dim]~...[/yellow dim]"
        if self.milestone:
            result += f" [purple]%{self.milestone}[/purple]"
        if self.assignees:
            result += f" [cyan]{' '.join(['@' + a for a in self.assignees])}[/cyan]"
        if self.description:
            result += " [white][...][/white]"
        return result

    @property
    def references(self) -> set[int]:
        # find all \b#\.(\d+)\b in description
        references = set()
        for word in self.description.strip().split(" "):
            if word.startswith("#.") and word[2:].strip().isdigit():
                references.add(int(word[2:].strip()))

        return references

    def resolve_references(self, resolution_map: dict[int, int], strict=False) -> 'Issue':
        resolved_description = self.description
        for reference in self.references:
            if (resolved := resolution_map.get(reference)):
                resolved_description = resolved_description.replace(f"#.{reference}", f"#{resolved}")
            elif strict:
                raise Exception(f"Could not resolve reference #.{reference}")

        return Issue(**(self._asdict() | {"description": resolved_description}))
            

    def submit(self, submitter_args: list[str]):
        remote_url = self._get_remote_url()
        if remote_url.hostname == "github.com":
            return self._github_submit(submitter_args)
        else:
            return self._gitlab_submit(submitter_args)

    def _get_remote_url(self):
        try:
            origin =  subprocess.run(
                ["git", "remote", "get-url", "origin"], capture_output=True
            ).stdout.decode()
            # fake an HTTPs URL from a SSH one
            if origin.startswith("git@"):
                origin = origin.replace(":", "/").replace("git@", "https://")
            return urlparse(origin)
        except subprocess.CalledProcessError as e:
            raise ValueError(
                "Could not determine remote url, make sure that you are inside of a git repository that has a remote named 'origin'"
            ) from e

    def _gitlab_submit(self, submitter_args: list[str]) -> int|None:
        command = ["glab", "issue", "new"]
        if self.title:
            command += ["-t", self.title]
        command += ["-d", self.description or ""]
        for a in self.assignees:
            command += ["-a", a if a != "me" else "@me"]
        for l in self.labels:
            command += ["-l", l]
        if self.milestone:
            command += ["-m", self.milestone]
        command.extend(submitter_args)
        out = self._run(command)
        # parse issue number from command output url: https://.+/-/issues/(\d+)
        if out and (url := re.search(r"https://.+/-/issues/(\d+)", out)):
            return int(url.group(1))
        
        # raise Exception(f"Could not parse issue number from {out!r}")

    def _github_submit(self, submitter_args: list[str]) -> int|None:
        command = ["gh", "issue", "new"]
        if self.title:
            command += ["-t", self.title]
        command += ["-b", self.description or ""]
        for a in self.assignees:
            command += ["-a", a if a != "me" else "@me"]
        for l in self.labels:
            command += ["-l", l]
        if self.milestone:
            command += ["-m", self.milestone]
        command.extend(submitter_args)
        out = self._run(command)
        # parse issue number from command output url: https://github.com/.+/issues/(\d+)
        pattern = re.compile(r"https:\/\/github\.com\/.+\/issues\/(\d+)")
        if out and (url := pattern.search(out)):
            return int(url.group(1))

        # raise Exception(f"Could not parse issue number from {out!r}, looked for regex {pattern}")
        return None

    def _run(self, command):
        if dry_running() or debugging():
            print(
                f"{'Would run' if dry_running() else 'Running'} [white bold]{subprocess.list2cmdline(command)}[/]"
            )
        if not dry_running():
            try:
                out = subprocess.run(command, check=True, capture_output=True)
                return out.stderr.decode() + "\n" + out.stdout.decode()
            except subprocess.CalledProcessError as e:
                print(
                    f"Calling [white bold]{e.cmd}[/] failed with code [white bold]{e.returncode}[/]:\n{NEWLINE.join(TAB + line for line in e.stderr.decode().splitlines())}"
                )

    @staticmethod
    def _word_and_sigil(raw_word: str) -> tuple[str, str]:
        if raw_word.startswith("#.") and raw_word[2:].isdigit():
            return "#.", raw_word[2:]

        sigil = raw_word[0]
        word = raw_word[1:]
        if sigil not in ("~", "%", "@"):
            sigil = ""
            word = raw_word
        return sigil, word

    # The boolean is true if the issue expects a description (ending ':')
    @classmethod
    def parse(cls, raw: str) -> tuple["Issue", bool]:
        raw = raw.strip()
        expects_description = False
        if raw.endswith(":"):
            expects_description = True
            raw = raw[:-1].strip()

        title = ""
        description = ""
        labels = set()
        assignees = set()
        milestone = ""
        reference = None
        # only labels/milestones/assignees at the beginning or end of the line are not added to the title as words
        add_to_title = False
        remaining_words = [word.strip() for word in raw.split(" ") if word.strip()]
        _debug_sigils = []

        while remaining_words:
            sigil, word = cls._word_and_sigil(remaining_words.pop(0))

            _debug_sigils.append(sigil)

            if sigil and add_to_title:
                title += f" {word}"

            match sigil:
                case "~":
                    labels.add(word)
                case "%":
                    milestone = word
                case "@":
                    assignees.add(word)
                case "#.":
                    reference = int(word)
                case _:
                    title += f" {word}"
                    # add to title if there are remaining regular words
                    add_to_title = any(
                        not sigil
                        for (sigil, _) in map(cls._word_and_sigil, remaining_words)
                    )

        return (
            cls(
                title=title.strip(),
                description=description,
                labels=labels,
                assignees=assignees,
                milestone=milestone,
                reference=reference 
            ),
            expects_description,
        )


def tree_to_text(tree: dict[str, Any], recursion_depth=0) -> str:
    result = ""
    for line, children in tree.items():
        result += TAB * recursion_depth + line.strip() + NEWLINE
        if children is not None:
            result += tree_to_text(children, recursion_depth + 1)
    return result


def parse_issue_fragment(
    issue_fragment: str,
    children: dict[str, Any],
    current_issue: Issue,
    recursion_depth=0,
    cli_options: dict[str, Any] | None = None,
) -> list[Issue]:
    if not cli_options:
        cli_options = {}
    log = lambda *args, **kwargs: print(
        f"[white]{issue_fragment[:50]: <50}[/white]\t{TAB*recursion_depth}",
        *args,
        **kwargs,
    )

    if issue_fragment.strip().startswith("//"):
        log(f"[yellow bold]Skipping comment[/]")
        return []
    log(f"Inheriting from {current_issue.display()}")
    current_title = current_issue.title
    current_description = current_issue.description
    current_labels = set(current_issue.labels)
    current_assignees = set(current_issue.assignees)
    current_milestone = current_issue.milestone

    parsed, expecting_description = Issue.parse(issue_fragment)
    if expecting_description:
        log(f"[white dim]{parsed} expects a description[/]")

    current_title = parsed.title
    current_labels |= parsed.labels
    current_assignees |= parsed.assignees
    if parsed.milestone:
        current_milestone = parsed.milestone
    if expecting_description:
        if children is None:
            raise ValueError(f"Expected a description after {issue_fragment!r}")
        current_description = tree_to_text(children, 0)

    current_issue = Issue(
        title=current_title,
        description=current_description,
        labels=current_labels,
        assignees=current_assignees,
        milestone=current_milestone,
        reference=parsed.reference
    )

    if current_issue.title:
        log(f"Made {current_issue.display()}")
        return [current_issue]

    if not expecting_description and children is not None:
        result = []
        log(f"Making children from {current_issue.display()}")
        for child, grandchildren in children.items():
            result.extend(
                parse_issue_fragment(
                    child,
                    grandchildren,
                    current_issue,
                    recursion_depth + 1,
                    cli_options,
                )
            )
        return result

    log(f"[red bold]Issue {issue_fragment!r} has no title and no children[/red bold]")
    return []


def parse(raw: str) -> Iterable[Issue]:
    for item in Node.to_dict(raw).items():
        debug(f"Processing {item!r}")
        for issue in parse_issue_fragment(*item, Issue("", "", set(), set(), "")):
            yield issue



from rich import print
import os


def debugging():
    return os.environ.get("ISSURGE_DEBUG")


def dry_running():
    return os.environ.get("ISSURGE_DRY_RUN")


def debug(*args, **kwargs):
    if os.environ.get("ISSURGE_DEBUG"):
        print(*args, **kwargs)


TAB = "\t"
NEWLINE = "\n"



from .parser import Issue


def create_issue(words: str) -> Issue:
    issue, expects_description = Issue.parse(words)
    if expects_description:
        print(
            "Please enter a description for the issue (submit 2 empty lines to finish):"
        )
        description = ""
        next_empty_is_finish = False
        while True:
            line = input("> ")
            if line == "":
                if next_empty_is_finish:
                    break
                next_empty_is_finish = True
            else:
                next_empty_is_finish = False

            description += line + "\n"

        issue = Issue(**(issue._asdict() | {"description": description.strip()}))

    return issue



import os
from unittest.mock import Mock

import pytest

import issurge.utils
from issurge.utils import debug, debugging, dry_running


def test_debugging_is_false_by_default():
    assert not debugging()


def test_debugging_is_true_when_issurge_debug_is_set():
    os.environ["ISSURGE_DEBUG"] = "1"
    assert debugging()
    del os.environ["ISSURGE_DEBUG"]


def test_dry_running_is_false_by_default():
    assert not dry_running()


def test_dry_running_is_true_when_issurge_dry_run_is_set():
    os.environ["ISSURGE_DRY_RUN"] = "1"
    assert dry_running()
    del os.environ["ISSURGE_DRY_RUN"]


def test_debug_only_prints_when_debugging_is_true():
    issurge.utils.print = Mock()
    os.environ["ISSURGE_DEBUG"] = "1"
    debug("debug")
    assert len(issurge.utils.print.mock_calls) == 1
    assert issurge.utils.print.mock_calls[0].args[0] == "debug"
    del os.environ["ISSURGE_DEBUG"]



from unittest.mock import Mock

import pytest

from issurge import interactive
from issurge.parser import Issue


@pytest.mark.parametrize(
    "words, expected",
    [
        ("", Issue()),
        ("a simple test right there", Issue(title="a simple test right there")),
        (
            "@me some ~labels to ~organize issues ~bug",
            Issue(
                title="some labels to organize issues",
                labels={"labels", "organize", "bug"},
                assignees={"me"},
            ),
        ),
        (
            "a %milestone to keep ~track of stuff",
            Issue(
                title="a milestone to keep track of stuff",
                labels={"track"},
                milestone="milestone",
            ),
        ),
    ],
)
def test_create_issue_from_words(words, expected):
    actual = interactive.create_issue(words)
    assert actual == expected


def test_create_issue_from_words_with_description():
    global called_count
    called_count = 0

    def mocked_input(prompt):
        global called_count
        called_count += 1
        return {
            1: "A description",
            2: "",
            3: "Another line",
            4: "With more content",
        }.get(called_count, "")

    interactive.input = Mock(wraps=mocked_input)
    actual = interactive.create_issue("A label with a description following it ~now:")
    expected = Issue(
        title="A label with a description following it",
        labels={"now"},
        description="A description\n\nAnother line\nWith more content",
    )
    assert actual == expected
    assert len(interactive.input.mock_calls) == 6



#!/usr/bin/env python
"""
Usage:
    issurge [options] new <words>...
    issurge [options] <file> [--] [<submitter-args>...]
    issurge --help

issurge new <words>... acts like echo <words>... | issurge /dev/stdin, but also asks for a description if the issue ends with `:'.

<submitter-args> contains arguments that will be passed as-is to the end of all `glab' commands

Options:
    --dry-run   Don't actually post the issues
    --debug     Print debug information
"""

import os
from pathlib import Path

from docopt import docopt
from rich import print

from issurge.parser import parse
from issurge.utils import debug, dry_running
from issurge import interactive


def run(opts=None):
    opts = opts or docopt(__doc__)
    os.environ["ISSURGE_DEBUG"] = "1" if opts["--debug"] else ""
    os.environ["ISSURGE_DRY_RUN"] = "1" if opts["--dry-run"] else ""

    debug(f"Running with options: {opts}")
    if opts["new"]:
        issue = interactive.create_issue(" ".join(opts["<words>"]))
        debug(f"Submitting {issue.display()}")
        number = issue.submit(opts["<submitter-args>"])
        print(f"Created issue #{number}")
    else:
        print("Submitting issues...")
        references_resolutions: dict[int, int] = {}
        for issue in parse(Path(opts["<file>"]).read_text()):
            issue = issue.resolve_references(references_resolutions, strict=not dry_running())
            number = issue.submit(opts["<submitter-args>"])
            print(f"Created issue #{number}")
            if issue.reference and number:
                references_resolutions[issue.reference] = number
